
/*
 * $FreeBSD$
 */

#include <bsnmp/snmpmod.h>

#include <sys/types.h>
#include <assert.h>
#include <errno.h>
#include <fcntl.h>
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <syslog.h>
#include <unistd.h>
#include <libutil.h>

#include "mysql.h"
#include "mysql_oid.h"
#include "mysql_tree.h"

struct lmodule *module;
static MYSQL mysql;
static char *_query (const char *str, int type);

int
mysql_master (struct snmp_context __unused *ctx, struct snmp_value *val,
        u_int sub, u_int __unused vindex, enum snmp_op op)
{
    asn_subid_t which = val->var.subs[sub - 1];

    if (op == SNMP_OP_SET)
        return (SNMP_ERR_NOT_WRITEABLE);

    if (op == SNMP_OP_GET)
    {
        switch (which)
        {
            case LEAF_mysqlMasterConnections:
                val->v.integer = atoi (_query ("Connections", 1));
                break;
            case LEAF_mysqlMasterSlowQuery:
                val->v.integer = atoi (_query ("SlowQuery", 1));
                break;
            case LEAF_mysqlMasterUptime:
                val->v.integer = atoi (_query ("Uptime", 1));
                break;
            case LEAF_mysqlMasterQuestions:
                val->v.integer = atoi (_query ("Questions", 1));
                break;
            default:
                return (SNMP_ERR_NOSUCHNAME);
        }

        return (SNMP_ERR_NOERROR);
    }

    abort ();
}

int
mysql_slave (struct snmp_context __unused *ctx, struct snmp_value *val,
        u_int sub, u_int __unused vindex, enum snmp_op op)
{
    asn_subid_t which = val->var.subs[sub - 1];

    if (op == SNMP_OP_SET)
        return (SNMP_ERR_NOT_WRITEABLE);

    if (op == SNMP_OP_GET)
    {
        switch (which)
        {
            default:
                return (SNMP_ERR_NOSUCHNAME);
        }

        return (SNMP_ERR_NOERROR);
    }

    abort ();
}

static
char *_query (const char *str, int type)
{
    MYSQL_RES *res = NULL;
    MYSQL_ROW row;
    char *query = NULL;
    int len, ret = 0;

    mysql_ping (&mysql);

    if (type)
    {
        len = strlen (str) + 35;
        query = (char *) malloc (len);
        snprintf (query, len, "SHOW STATUS WHERE Variable_name='%s'", str);
    }
    else
    {
        len = strlen (str) + 41;
        query = (char *) malloc (len);
        snprintf (query, len, "SHOW SLAVE STATUS WHERE Variable_name='%s'", str);
    }

    syslog(LOG_ERR, "_query: %s", query);
    ret = mysql_query (&mysql, query);
    free (query);
    if (ret != 0)
        return "";

    res = mysql_store_result (&mysql);
    if (mysql_num_rows (res) != 1)
    {
        mysql_free_result (res);
        return "";
    }

    row = mysql_fetch_row (res);
    mysql_free_result (res);

    return row[1];
}

static int
my_init (struct lmodule *mod, int __unused argc, char __unused *argv[])
{
    module = mod;
    int fd;
    const char *p;
    char *username, *password;
    my_bool reconnect = 0;
    properties prop;

    fd = open (CONF_FILE, O_RDONLY);

    /* could not open config file, set as default. */
    if (fd < 0)
    {
        username = strdup ("root");
        password = strdup ("");
    }
    else
    {
        prop = properties_read (fd);
        p = property_find (prop, "username");
        if (p == NULL)
        {
            syslog (LOG_ERR, "Minssing \"username\" property");
            return (-1);
        }
        username = strdup (p);

        p = property_find (prop, "password");
        if (p == NULL)
        {
            syslog (LOG_ERR, "Minssing \"password\" property");
            return (-1);
        }
        password = strdup (p);

        properties_free (prop);
        close (fd);
    }

    mysql_init (&mysql);
#if MYSQL_VERSION_ID < 50000
    mysql.reconnect= reconnect;
#else
    mysql_options (&mysql, MYSQL_OPT_RECONNECT, (char *)&reconnect);
#endif

    if (!mysql_real_connect (&mysql, "localhost", username, password, "mysql", 0, NULL, 0))
    {
        syslog (LOG_ERR, "my_init(): %s", mysql_error (&mysql));
        return (-1);
    }

    return (0);
}

static int
my_fini (void)
{
    mysql_close (&mysql);

    return (0);
}

const struct snmp_module config = {
    .comment = "This module implements a MIB for the mysql server status.",
    .init = my_init,
    .fini = my_fini,
    .tree = mysql_ctree,
    .tree_size = mysql_CTREE_SIZE,
};
